def transform(dataset, resampling_factor=[1, 1, 1]):
    """Resample dataset"""

    from tomviz import utils
    import scipy.ndimage
    import numpy as np

    array = dataset.active_scalars

    # Transform the dataset.
    result_shape = utils.zoom_shape(array, resampling_factor)
    result = np.empty(result_shape, array.dtype, order='F')
    scipy.ndimage.interpolation.zoom(array, resampling_factor, output=result)

    # Set the result as the new scalars.
    dataset.active_scalars = result

    # Update tilt angles if dataset is a tilt series.
    if resampling_factor[2] != 1:
        try:
            tilt_angles = dataset.tilt_angles

            result_shape = utils.zoom_shape(tilt_angles, resampling_factor[2])
            result = np.empty(result_shape, array.dtype, order='F')
            scipy.ndimage.interpolation.zoom(
                tilt_angles, resampling_factor[2], output=result)
            dataset.tilt_angles = result
        except: # noqa
            # TODO What exception are we ignoring?
            pass



# Perform alignment to the estimated rotation axis
#
# Developed as part of the tomviz project (www.tomviz.com).


def transform(dataset, SHIFT=None, rotation_angle=90.0, tilt_axis=0):
    from tomviz import utils
    from scipy import ndimage
    import numpy as np

    data_py = dataset.active_scalars # Get data as numpy array.

    if data_py is None: #Check if data exists
        raise RuntimeError("No data array found!")
    if SHIFT is None:
        SHIFT = np.zeros(len(data_py.shape), dtype=np.int)
    data_py_return = np.empty_like(data_py)
    ndimage.interpolation.shift(data_py, SHIFT, order=0, output=data_py_return)

    rotation_axis = 2 # This operator always assumes the rotation axis is Z
    if rotation_angle == []: # If tilt angle not given, assign it to 90 degrees.
        rotation_angle = 90

    axis1 = (rotation_axis + 1) % 3
    axis2 = (rotation_axis + 2) % 3
    axes = (axis1, axis2)
    shape = utils.rotate_shape(data_py_return, rotation_angle, axes=axes)
    data_py_return2 = np.empty(shape, data_py_return.dtype, order='F')
    ndimage.interpolation.rotate(
        data_py_return, rotation_angle, output=data_py_return2, axes=axes)

    dataset.active_scalars = data_py_return2



import collections
import copy
import h5py
import importlib
import os
import numpy as np
import logging
import tempfile
import socket
import abc
import stat
import json
import six
import errno

from tqdm import tqdm

from tomviz._internal import find_transform_function

from tomviz.external_dataset import Dataset

LOG_FORMAT = '[%(asctime)s] %(levelname)s: %(message)s'

logger = logging.getLogger('tomviz')
logger.setLevel(logging.INFO)
stream_handler = logging.StreamHandler()
formatter = logging.Formatter(LOG_FORMAT)
stream_handler.setFormatter(formatter)
logger.addHandler(stream_handler)

DIMS = ['dim1', 'dim2', 'dim3']
ANGLE_UNITS = [b'[deg]', b'[rad]']

Dim = collections.namedtuple('Dim', 'path values name units')


class ProgressBase(object):
    def started(self, op=None):
        self._operator_index = op

    def finished(self, op=None):
        pass


class TqdmProgress(ProgressBase):
    """
    Class used to update operator progress. This replaces the C++ bound one that
    is used inside the Qt applicaition.
    """

    def __init__(self):
        self._maximum = None
        self._value = None
        self._message = None
        self._progress_bar = None

    @property
    def maximum(self):
        """
        Property defining the maximum progress value
        """
        return self._maximum

    @maximum.setter
    def maximum(self, value):
        self._progress_bar = tqdm(total=value)
        self._maximum = value

    @property
    def value(self):
        """
        Property defining the current progress value
        """
        return self._value

    @value.setter
    def value(self, value):
        """
        Updates the progress of the the operator.

        :param value The current progress value.
        :type value: int
        """
        self._progress_bar.update(value - self._progress_bar.n)
        self._value = value

    @property
    def message(self):
        """
        Property defining the current progress message
        """
        return self._message

    @message.setter
    def message(self, msg):
        """
        Updates the progress message of the the operator.

        :param msg The current progress message.
        :type msg: str
        """
        self._progress_bar.set_description(msg)
        self._message = msg

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        if self._progress_bar is not None:
            self._progress_bar.close()

        return False

    def finished(self, op=None):
        if self._progress_bar is not None:
            self._progress_bar.close()


class JsonProgress(ProgressBase, metaclass=abc.ABCMeta):
    """
    Abstract class used to update operator progress using JSON based messages.
    """

    @abc.abstractmethod
    def write(self, data):
        """
        Method the write JSON progress message. Implemented by subclass.
        """

    @abc.abstractmethod
    def write_to_file(self, data):
        """
        Write data to write and return path. Implemented by subclass.
        """

    def set_operator_index(self, index):
        self._operator_index = index

    @property
    def maximum(self):
        """
        Property defining the maximum progress value
        """
        return self._maximum

    @maximum.setter
    def maximum(self, value):
        msg = {
            'type': 'progress.maximum',
            'operator': self._operator_index,
            'value': value
        }
        self.write(msg)
        self._maximum = value

    @property
    def value(self):
        """
        Property defining the current progress value
        """
        return self._value

    @value.setter
    def value(self, value):
        """
        Updates the progress of the the operator.

        :param value The current progress value.
        :type value: int
        """
        msg = {
            'type': 'progress.step',
            'operator': self._operator_index,
            'value': value
        }
        self.write(msg)

        self._value = value

    @property
    def message(self):
        """
        Property defining the current progress message
        """
        return self._message

    @message.setter
    def message(self, msg):
        """
        Updates the progress message of the the operator.

        :param msg The current progress message.
        :type msg: str
        """
        m = {
            'type': 'progress.message',
            'operator': self._operator_index,
            'value': msg
        }
        self.write(m)

        self._message = msg

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, value):
        """
        Updates the progress of the the operator.

        :param data The current progress data value.
        :type value: numpy.ndarray
        """
        path = self.write_to_file(value)
        msg = {
            'type': 'progress.data',
            'operator': self._operator_index,
            'value': path
        }
        self.write(msg)
        self._data = value

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        return False

    def started(self, op=None):
        super(JsonProgress, self).started(op)
        msg = {
            'type': 'started'
        }

        if op is not None:
            msg['operator'] = op

        self.write(msg)

    def finished(self, op=None):
        super(JsonProgress, self).started(op)
        msg = {
            'type': 'finished'
        }

        if op is not None:
            msg['operator'] = op

        self.write(msg)


class WriteToFileMixin(object):
    def write_to_file(self, dataobject):
        filename = '%d.emd' % self._sequence_number
        path = os.path.join(os.path.dirname(self._path), filename)
        _write_emd(path, dataobject)
        self._sequence_number += 1

        return filename


class LocalSocketProgress(WriteToFileMixin, JsonProgress):
    """
    Class used to update operator progress. Connects to QLocalServer and writes
    JSON message updating the UI on pipeline progress.
    """

    def __init__(self, socket_path):
        self._maximum = None
        self._value = None
        self._message = None
        self._connection = None
        self._path = socket_path
        self._sequence_number = 0

        try:
            mode = os.stat(self._path).st_mode
            if stat.S_ISSOCK(mode):
                self._uds_connect()
            else:
                raise Exception('Invalid progress path type.')

        except OSError:
            raise Exception('TqdmProgress path doesn\'t exist')

    def _uds_connect(self):
        self._connection = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self._connection.connect(self._path)

    def write(self, data):
        data = ('%s\n' % json.dumps(data)).encode('utf8')
        if isinstance(self._connection, socket.socket):
            self._connection.sendall(data)
        else:
            self._connection.write(data)

    def __exit__(self, *exc):
        if self._connection is not None:
            self._connection.close()

        return False


class FilesProgress(WriteToFileMixin, JsonProgress):
    """
    Class used to update operator progress by writing a sequence of files to a
    directory.
    """
    def __init__(self, path):
        self._path = path
        self._sequence_number = 0

    def write(self, data):
        file_path = os.path.join(self._path,
                                 'progress%d' % self._sequence_number)
        self._sequence_number += 1

        with open(file_path, 'w') as f:
            json.dump(data, f)

        return filename


def _progress(progress_method, progress_path):
    if progress_method == 'tqdm':
        return TqdmProgress()
    elif progress_method == 'socket':
        return LocalSocketProgress(progress_path)
    elif progress_method == 'files':
        return FilesProgress(progress_path)
    else:
        raise Exception('Unrecognized progress method: %s' % progress_method)


class OperatorWrapper(object):
    canceled = False


def _load_operator_module(label, script):
    # Load the operator module, we write the code to a temporary file before
    # using importlib to do the loading ( couldn't figure a way to directly
    # load a module from a string).
    fd = None
    path = None

    try:
        (fd, path) = tempfile.mkstemp(suffix='.py', text=True)
        os.write(fd, script.encode())
        os.close(fd)
        fd = None

        spec = importlib.util.spec_from_file_location(label, path)
        operator_module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(operator_module)

    finally:
        if fd is not None:
            os.close(fd)
        if path is not None:
            os.unlink(path)

    return operator_module


def _read_dataset(dataset, options=None):
    if options is None or 'subsampleSettings' not in options:
        return dataset[:]

    strides = options.get('subsampleSettings', {}).get('strides', [1] * 3)
    bnds = options.get('subsampleSettings', {}).get('volumeBounds', [-1] * 6)

    if len(bnds) != 6 or any(x < 0 for x in bnds):
        # Set the default bounds
        for i in range(3):
            bnds[i * 2] = 0
            bnds[i * 2 + 1] = dataset.shape[i]

    # These slice specifications are forwarded directly to the HDF5
    # "hyperslab" selections, and it loads only the data requested.
    return dataset[bnds[0]:bnds[1]:strides[0],
                   bnds[2]:bnds[3]:strides[1],
                   bnds[4]:bnds[5]:strides[2]]


def _swap_dims(dims, i, j):
    # Keeps the paths the same, but swaps everything else
    tmp = Dim(dims[i].path, dims[j].values, dims[j].name, dims[j].units)
    dims[j] = Dim(dims[j].path, dims[i].values, dims[i].name, dims[i].units)
    dims[i] = tmp


def _read_emd(path, options=None):
    with h5py.File(path, 'r') as f:
        tomography = f['data/tomography']

        dims = []
        for dim in DIMS:
            dims.append(Dim(dim,
                            tomography[dim][:],
                            tomography[dim].attrs['name'][0],
                            tomography[dim].attrs['units'][0]))

        data = tomography['data']
        # We default the name to ImageScalars
        name = data.attrs.get('name', 'ImageScalars')
        if not isinstance(name, str):
            name = name[0].decode()

        arrays = [(name, _read_dataset(data, options))]

        # See if we have any additional channels
        tomviz_scalars = tomography.get('tomviz_scalars')
        if isinstance(tomviz_scalars, h5py.Group):
            # Only get hard linked datasets, since there is likely a
            # soft link pointing to "/data/tomography/data".
            def is_hard_link(name):
                link = tomviz_scalars.get(name, getlink=True)
                return isinstance(link, h5py.HardLink)

            keys = list(filter(is_hard_link, tomviz_scalars.keys()))
            # Get the datasets
            channel_datasets = [
                (key, _read_dataset(tomviz_scalars[key], options))
                for key in keys
            ]
            arrays += channel_datasets

        # If this is a tilt series, swap the X and Z axes
        tilt_axis = None
        if dims[0].name == b'angles' or dims[0].units in ANGLE_UNITS:
            arrays = [(name, np.transpose(data, [2, 1, 0])) for (name, data)
                      in arrays]

            # Swap the first and last dims
            _swap_dims(dims, 0, -1)
            tilt_axis = 2

        # EMD stores data as row major order.  VTK expects column major order.
        arrays = [(name, np.asfortranarray(data)) for (name, data) in arrays]

        output = {
            'arrays': arrays,
            'dims': dims,
            'tilt_axis': tilt_axis
        }

        if dims is not None and dims[-1].name == b'angles':
            output['tilt_angles'] = dims[-1].values[:].astype(np.float64)

        return output


def _get_arrays_for_writing(dataset):
    tilt_angles = dataset.tilt_angles
    tilt_axis = dataset.tilt_axis
    active_array = dataset.active_scalars

    # Separate out the extra channels/arrays as we store them separately
    extra_arrays = {name: array for name, array
                    in dataset.arrays.items()
                    if id(array) != id(active_array)}

    # If this is a tilt series, swap the X and Z axes
    if tilt_angles is not None and tilt_axis == 2:
        active_array = np.transpose(active_array, [2, 1, 0])
        extra_arrays = {name: np.transpose(array, [2, 1, 0]) for (name, array)
                        in extra_arrays.items()}

    # Switch back to row major order for EMD stores
    active_array = np.ascontiguousarray(active_array)
    extra_arrays = {name: np.ascontiguousarray(array) for (name, array)
                    in extra_arrays.items()}

    # We can't do 16 bit floats, so up the size if they are 16 bit
    if active_array.dtype == np.float16:
        active_array = active_array.astype(np.float32)
    for key, value in extra_arrays.items():
        if value.dtype == np.float16:
            extra_arrays[key] = value.astype(np.float32)

    return active_array, extra_arrays


def _get_dims_for_writing(dataset, data, default_dims=None):
    tilt_angles = dataset.tilt_angles
    tilt_axis = dataset.tilt_axis
    spacing = dataset.spacing

    if default_dims is None:
        # Set the default dims
        dims = []
        names = ['x', 'y', 'z']

        for i, name in enumerate(names):
            values = np.array(range(data.shape[i]))
            dims.append(Dim('dim%d' % (i+1), values, name, '[n_m]'))
    else:
        # Make a deep copy to modify
        dims = copy.deepcopy(default_dims)

    # Override the dims if spacing is set
    if spacing is not None:
        for i, dim in enumerate(dims):
            end = data.shape[i] * spacing[i]
            res = np.arange(0, end, spacing[i])
            dims[i] = Dim(dim.path, res, dim.name, dim.units)

    if tilt_angles is not None:
        if tilt_axis == 2:
            # Swap the first and last dims
            _swap_dims(dims, 0, -1)

        units = dims[0].units if dims[0].units in ANGLE_UNITS else '[deg]'

        # Make the x axis the tilt angles
        dims[0] = Dim(dims[0].path, tilt_angles, 'angles', units)
    else:
        # In case the input was a tilt series, make the first dim x,
        # and the units [n_m]
        if dims[0].name == 'angles':
            dims[0].name = 'x'
            dims[0].units = '[n_m]'

    return dims


def _write_emd(path, dataset, dims=None):
    active_array, extra_arrays = _get_arrays_for_writing(dataset)

    with h5py.File(path, 'w') as f:
        f.attrs.create('version_major', 0, dtype='uint32')
        f.attrs.create('version_minor', 2, dtype='uint32')
        data_group = f.create_group('data')
        tomography_group = data_group.create_group('tomography')
        tomography_group.attrs.create('emd_group_type', 1, dtype='uint32')
        data = tomography_group.create_dataset('data', data=active_array)
        data.attrs['name'] = np.string_(dataset.active_name)

        dims = _get_dims_for_writing(dataset, data, dims)

        # add dimension vectors
        for dim in dims:
            d = tomography_group.create_dataset(dim.path, dim.values.shape)
            d.attrs['name'] = np.string_(dim.name)
            d.attrs['units'] = np.string_(dim.units)
            d[:] = dim.values

        # If we have extra scalars add them under tomviz_scalars
        tomviz_scalars = tomography_group.create_group('tomviz_scalars')
        if extra_arrays:
            for (name, array) in extra_arrays.items():
                tomviz_scalars.create_dataset(name, data=array)

        # Create a soft link to the active array
        active_name = dataset.active_name
        tomviz_scalars[active_name] = h5py.SoftLink('/data/tomography/data')


def _read_data_exchange(path, options=None):
    with h5py.File(path, 'r') as f:
        g = f['/exchange']

        dataset_names = ['data', 'data_dark', 'data_white', 'theta']
        datasets = {}
        for name in dataset_names:
            if name in g:
                # Don't read theta with subsample options
                if name == 'theta':
                    datasets[name] = g[name][:]
                    continue

                datasets[name] = _read_dataset(g[name], options)

        # Data Exchange stores data as row major order.
        # VTK expects column major order.
        if options is None:
            to_fortran = True
        else:
            to_fortran = not options.get('keep_c_ordering', False)

        if to_fortran:
            datasets = {name: np.asfortranarray(data)
                        for (name, data) in datasets.items()}

            if 'theta' in datasets:
                # Swap x and z axes
                swap_keys = ['data', 'data_dark', 'data_white']
                for key in swap_keys:
                    datasets[key] = np.transpose(datasets[key], [2, 1, 0])

        tilt_axis = None
        if 'theta' in datasets:
            tilt_axis = 2 if to_fortran else 0

        output = {
            'arrays': [('data', datasets.get('data'))],
            'data_dark': datasets.get('data_dark'),
            'data_white': datasets.get('data_white'),
            'tilt_angles': datasets.get('theta'),
            'tilt_axis': tilt_axis
        }

        return output


def _is_data_exchange(path):
    # It must have an HDF5 extension
    exts = ['.h5', '.hdf5']
    if not any([path.suffix.lower() == x for x in exts]):
        return False

    # Open it up and make sure /exchange/data exists
    with h5py.File(path, 'r') as f:
        return '/exchange/data' in f


def _execute_transform(operator_label, transform, arguments, input, progress):
    # Update the progress attribute to an instance that will work outside the
    # application and give use a nice progress information.
    spinner = None
    supports_progress = hasattr(transform, '__self__')
    if supports_progress:
        transform.__self__.progress = progress

        # Stub out the operator wrapper
        transform.__self__._operator_wrapper = OperatorWrapper()

    logger.info('Executing \'%s\' operator' % operator_label)
    if not supports_progress:
        print('Operator doesn\'t support progress updates.')

    result = None
    # Special cases for marking as volume or tilt series
    if operator_label == 'ConvertToVolume':
        # Easy peasy
        input.tilt_angles = None
    elif operator_label == 'SetTiltAngles':
        # Set the tilt angles so downstream operator can retrieve them
        # arguments the tilt angles.
        input.tilt_angles = np.array(arguments['angles']).astype(np.float64)
    else:
        # Now run the operator
        result = transform(input, **arguments)
    if spinner is not None:
        update_spinner.cancel()
        spinner.finish()

    return result


def _load_transform_functions(operators):
    transform_functions = []
    for operator in operators:
        # Special case for marking as volume or tilt series
        if operator['type'] == 'ConvertToVolume':
            transform_functions.append((operator['type'], None, {}))
            continue
        elif operator['type'] == 'SetTiltAngles':
            # Just save the angles
            angles = {'angles': [float(a) for a in operator['angles']]}
            transform_functions.append((operator['type'], None, angles))
            continue

        if 'script' not in operator:
            raise Exception(
                'No script property found. C++ operator are not supported.')

        operator_script = operator['script']
        operator_label = operator['label']

        operator_module = _load_operator_module(operator_label, operator_script)
        transform = find_transform_function(operator_module)

        # partial apply the arguments
        arguments = {}
        if 'arguments' in operator:
            arguments = operator['arguments']

        transform_functions.append((operator_label, transform, arguments))

    return transform_functions


def _write_child_data(result, operator_index, output_file_path, dims):
    for (label, dataobject) in six.iteritems(result):
        # Only need write out data if the operator made updates.
        output_path = '.'
        if output_file_path is not None:
            output_path = os.path.dirname(output_file_path)

        # Make a directory with the operator index
        operator_path = os.path.join(output_path, str(operator_index))
        try:
            stat_result = os.stat(output_path)
            os.makedirs(operator_path)
            # We need to chown to the user and group who owns the output
            # path ( for docker execution )
            os.chown(operator_path, stat_result.st_uid, stat_result.st_gid)
        except OSError as exc:
            if exc.errno == errno.EEXIST and os.path.isdir(operator_path):
                pass
            else:
                raise

        # Now write out the data
        child_data_path = os.path.join(operator_path, '%s.emd' % label)
        _write_emd(child_data_path, dataobject, dims)


def execute(operators, start_at, data_file_path, output_file_path,
            progress_method, progress_path, read_options=None):

    if _is_data_exchange(data_file_path):
        output = _read_data_exchange(data_file_path, read_options)
    else:
        # Assume it is emd
        output = _read_emd(data_file_path, read_options)

    arrays = output['arrays']
    dims = output.get('dims')

    # The first is the active array
    (active_array, _) = arrays[0]
    # Create dict of arrays
    arrays = {name: array for (name, array) in arrays}

    data = Dataset(arrays, active_array)
    if 'data_dark' in output:
        data.dark = output['data_dark']
    if 'data_white' in output:
        data.white = output['data_white']
    if 'tilt_angles' in output:
        data.tilt_angles = output['tilt_angles']
    if 'tilt_axis' in output:
        data.tilt_axis = output['tilt_axis']
    if dims is not None:
        # Convert to native type, as is required by itk
        data.spacing = [float(d.values[1] - d.values[0]) for d in dims]

    operators = operators[start_at:]
    transforms = _load_transform_functions(operators)
    with _progress(progress_method, progress_path) as progress:
        progress.started()
        operator_index = start_at
        for (label, transform, arguments) in transforms:
            progress.started(operator_index)
            result = _execute_transform(label, transform,
                                        arguments, data,
                                        progress)

            # Do we have any child data sources we need to write out?
            if result is not None:
                _write_child_data(result, operator_index,
                                  output_file_path, dims)

            progress.finished(operator_index)
            operator_index += 1

        logger.info('Execution complete.')
        # Now write out the transformed data.
        logger.info('Writing transformed data.')
        if output_file_path is None:
            output_file_path = '%s_transformed.emd' % \
                os.path.splitext(os.path.basename(data_file_path))[0]

        if result is None:
            _write_emd(output_file_path, data, dims)
        else:
            [(_, child_data)] = result.items()
            _write_emd(output_file_path, child_data, dims)
        logger.info('Write complete.')
        progress.finished()


if __name__ == '__main__':
    main()



import base64
import json
import os
import shutil
import zipfile
import tempfile

from paraview import simple
from paraview.web.dataset_builder import ImageDataSetBuilder
from paraview.web.dataset_builder import CompositeDataSetBuilder
from paraview.web.dataset_builder import VTKGeometryDataSetBuilder

DATA_DIRECTORY = 'data'
HTML_FILENAME = 'tomviz.html'
JS_FILENAME = 'tomviz.js'
HTML_WITH_DATA_FILENAME = 'tomviz_data.html'
DATA_FILENAME = 'data.tomviz'

arrayTypesMapping = '  bBhHiIlLfd'
jsMapping = {
    'b': 'Int8Array',
    'B': 'Uint8Array',
    'h': 'Int16Array',
    'H': 'Int16Array',
    'i': 'Int32Array',
    'I': 'Uint32Array',
    'l': 'Int32Array',
    'L': 'Uint32Array',
    'f': 'Float32Array',
    'd': 'Float64Array'
}


def web_export(*args, **kwargs):
    temp_dir = tempfile.mkdtemp()
    try:
        # Expecting only kwargs
        keepData = kwargs['keepData']
        executionPath = kwargs['executionPath']
        exportType = kwargs['exportType']
        htmlFilePath = kwargs['htmlFilePath']
        if not htmlFilePath.lower().endswith(('.html', '.htm')):
            htmlFilePath += '.html'

        dataFileName = '%s.zip' % os.path.splitext(
            os.path.basename(htmlFilePath))[0]
        dataFilePath = os.path.join(os.path.dirname(htmlFilePath), dataFileName)

        # Camera properties
        nbPhi = kwargs['nbPhi']
        nbTheta = kwargs['nbTheta']

        # Destination directory for data
        dest = '%s/data' % temp_dir

        # Extract initial setting for view
        view = simple.GetRenderView()
        viewState = {}
        for prop in ['CameraViewUp', 'CameraPosition', 'ViewSize']:
            viewState[prop] = tuple(view.GetProperty(prop).GetData())

        # Camera handling
        deltaPhi = int(360 / nbPhi)
        deltaTheta = int(180 / nbTheta)
        thetaMax = deltaTheta
        while thetaMax + deltaTheta < 90:
            thetaMax += deltaTheta
        camera = {
            'type': 'spherical',
            'phi': range(0, 360, deltaPhi),
            'theta': range(-thetaMax, thetaMax + 1, deltaTheta)
        }

        # Choose export mode:
        if exportType == 0:
            export_images(dest, camera, **kwargs)

        if exportType == 1:
            export_volume_exploration_images(dest, camera, **kwargs)

        if exportType == 2:
            export_contour_exploration_images(dest, camera, **kwargs)

        if exportType == 3:
            export_contours_geometry(dest, **kwargs)

        if exportType == 4:
            export_contour_exploration_geometry(dest, **kwargs)

        if exportType == 5:
            export_volume(dest, **kwargs)

        # Setup application
        copy_viewer(temp_dir, executionPath)

        # Compress only geometry data
        bundleDataToHTML(temp_dir, htmlFilePath,
                         dataFilePath=dataFilePath if keepData else None,
                         compress=exportType > 2)

        # Restore initial parameters
        for prop in viewState:
            view.GetProperty(prop).SetData(viewState[prop])
    finally:
        shutil.rmtree(temp_dir, ignore_errors=True)
# -----------------------------------------------------------------------------
# Helpers
# -----------------------------------------------------------------------------


def bundleDataToHTML(temp_dir, htmlFilePath, dataFilePath=None, compress=False):
    compression_type = zipfile.ZIP_DEFLATED if compress else zipfile.ZIP_STORED
    dataDir = os.path.join(temp_dir, DATA_DIRECTORY)
    srcHtmlPath = os.path.join(temp_dir, HTML_FILENAME)
    webResources = ['<style>.webResource { display: none; }</style>']

    if os.path.exists(dataDir):
        for dirName, subdirList, fileList in os.walk(dataDir):
            for fname in fileList:
                fullPath = os.path.join(dirName, fname)
                filePath = os.path.relpath(fullPath, dataDir)
                relPath = '%s/%s' % (DATA_DIRECTORY, filePath)
                content = ''

                if fname.endswith('.json'):
                    with open(fullPath, 'r', encoding='utf8') as data:
                        content = data.read()
                else:
                    with open(fullPath, 'rb') as data:
                        dataContent = data.read()
                        content = base64.b64encode(dataContent)
                        content = content.decode().replace('\n', '')

                webResources.append(
                    '<div class="webResource" data-url="%s">%s</div>'
                    % (relPath, content))

        webResources.append('<script>ready()</script></body>')

        # Create new output file
        with open(srcHtmlPath, mode='r', encoding='utf8') as srcHtml:
            with open(htmlFilePath, mode='w', encoding='utf8') as dstHtml:
                for line in srcHtml:
                    if '</body>' in line:
                        for webResource in webResources:
                            dstHtml.write(webResource)
                    else:
                        dstHtml.write(line)

        # Generate zip file for the data
        if dataFilePath is not None and os.path.exists(dataDir):
            with zipfile.ZipFile(dataFilePath, mode='w') as zf:
                for dirName, subdirList, fileList in os.walk(dataDir):
                    for fname in fileList:
                        fullPath = os.path.join(dirName, fname)
                        filePath = os.path.relpath(fullPath, dataDir)
                        relPath = '%s/%s' % (DATA_DIRECTORY, filePath)
                        zf.write(fullPath, arcname=relPath,
                                 compress_type=compression_type)


def get_proxy(id):
    session = simple.servermanager.ActiveConnection.Session
    remoteObj = session.GetRemoteObject(int(id))
    return simple.servermanager._getPyProxy(remoteObj)


def copy_viewer(destinationPath, executionPath):
    searchPath = executionPath
    for upDirTry in range(4):
        searchPath = os.path.normpath(os.path.join(searchPath, '..'))
        for root, dirs, files in os.walk(searchPath):
            if HTML_FILENAME in files and root != destinationPath:
                srcHtmlFile = os.path.join(root, HTML_FILENAME)
                srcJsFile = os.path.join(root, JS_FILENAME)
                dstHtmlFile = os.path.join(destinationPath, HTML_FILENAME)
                with open(srcHtmlFile, mode='r', encoding='utf8') as srcHTML, \
                        open(srcJsFile, mode='r', encoding='utf8') as srcJS, \
                        open(dstHtmlFile, mode='w', encoding='utf8') as dstHTML:
                    for line in srcHTML:
                        if '</body>' in line:
                            dstHTML.write('<script type="text/javascript">\n')
                            dstHTML.write(srcJS.read())
                            dstHTML.write('\n</script>\n</body>\n')
                        else:
                            dstHTML.write(line)
                    return


def add_scene_item(scene, name, proxy, view):
    hasNormal = False
    hasColor = False
    colors = {}
    representation = {}
    rep = simple.GetRepresentation(proxy, view)

    # Skip hidden object or volume
    if not rep.Visibility or rep.Representation == 'Volume':
        return

    for prop in ['Representation']:
        representation[prop] = rep.GetProperty(prop).GetData()

    pdInfo = proxy.GetPointDataInformation()
    numberOfPointArrays = pdInfo.GetNumberOfArrays()
    for idx in range(numberOfPointArrays):
        array = pdInfo.GetArray(idx)
        rangeValues = array.GetRange(-1)
        if array.Name == 'Normals':
            hasNormal = True
        if array.Name not in ['vtkValidPointMask', 'Normals']:
            hasColor = True
            if rangeValues[0] == rangeValues[1]:
                colors[array.Name] = {
                    'constant': rangeValues[0],
                    'location': 'POINT_DATA'
                }
            else:
                colors[array.Name] = {
                    'location': 'POINT_DATA',
                    'range': [i for i in rangeValues]
                }

    # Get information about cell data arrays
    cdInfo = proxy.GetCellDataInformation()
    numberOfCellArrays = cdInfo.GetNumberOfArrays()
    for idx in range(numberOfCellArrays):
        array = cdInfo.GetArray(idx)
        hasColor = True
        colors[array.Name] = {
            'location': 'CELL_DATA',
            'range': array.GetRange(-1)
        }

    # Make sure Normals are available if lighting by normals
    source = proxy
    if not hasColor or rep.Representation == 'Outline':
        colors = {'solid': {'constant': 0, 'location': 'POINT_DATA'}}
    elif 'light' in scene and 'normal' in scene['light'] and not hasNormal:
        rep.Visibility = 0
        surface = simple.ExtractSurface(Input=proxy)
        surfaceWithNormals = simple.GenerateSurfaceNormals(Input=surface)
        source = surfaceWithNormals

    scene['scene'].append({
        'name': name,
        'source': source,
        'colors': colors,
        'representation': representation
    })


def patch_data_range(destinationPath):
    originalPath = os.path.join(destinationPath, 'index.json')
    tmpPath = os.path.join(destinationPath, 'index_tmp.json')
    os.rename(originalPath, tmpPath)
    with open(originalPath, 'w', encoding='utf8') as outputJSON:
        with open(tmpPath, 'r', encoding='utf8') as inputJSON:
            content = json.loads(inputJSON.read())
            # Patch geometry range
            for fieldName in content['Geometry']['ranges']:
                content['Geometry']['ranges'][fieldName] = [0, 255]
            outputJSON.write(json.dumps(content, indent=2))
    os.remove(tmpPath)


def get_volume_piecewise(view):
    renderer = view.GetClientSideObject().GetRenderer()
    for volume in renderer.GetVolumes():
        if volume.GetClassName() == 'vtkVolume':
            return volume.GetProperty().GetScalarOpacity()
    return None


def build_control_points(xrgbArray):
    minValue = xrgbArray[0]
    maxValue = xrgbArray[-4]
    dataRange = maxValue - minValue
    controlpoints = []
    offset = 0
    while offset < len(xrgbArray):
        controlpoints.append({
            'x': (xrgbArray[offset] - minValue) / dataRange,
            'r': xrgbArray[offset + 1],
            'g': xrgbArray[offset + 2],
            'b': xrgbArray[offset + 3],
        })
        offset += 4
    return controlpoints


def get_volume_lookuptable_section(view):
    renderer = view.GetClientSideObject().GetRenderer()
    for volume in renderer.GetVolumes():
        if volume.GetClassName() == 'vtkVolume':
            array = volume.GetMapper().GetInput().GetPointData().GetScalars()
            fileName = array.GetName()
            lut = volume.GetProperty().GetRGBTransferFunction()
            controlpoints = []
            tupleHolder = list(range(6))
            for i in range(lut.GetSize()):
                lut.GetNodeValue(i, tupleHolder)
                controlpoints.append({
                    'x': tupleHolder[0],
                    'r': tupleHolder[1],
                    'g': tupleHolder[2],
                    'b': tupleHolder[3],
                })

            # Need to rescale x between [0, 1] for pvw
            minValue = controlpoints[0]['x']
            maxValue = controlpoints[-1]['x']
            dataRange = maxValue - minValue
            for node in controlpoints:
                node['x'] = (node['x'] - minValue) / dataRange

            return {fileName: {'controlpoints': controlpoints}}
    return {}


def get_source_lookuptable_section(source):
    luts = {}
    rep = simple.GetRepresentation(source)
    name = rep.ColorArrayName[1]
    lut = rep.LookupTable
    if lut:
        controlpoints = build_control_points(lut.RGBPoints)
        luts[name] = {'controlpoints': controlpoints}

    return luts


def get_contour():
    for key, value in simple.GetSources().items():
        if 'FlyingEdges' in key[0]:
            return value
        if 'Contour' in key[0]:
            return value
    return None


def get_trivial_producer():
    for key, value in simple.GetSources().items():
        if 'TrivialProducer' in key[0]:
            return value
    return None


def array_sampler(srcDims, dstDims, scale, inputArray):
    if scale == 1:
        return inputArray

    newSize = dstDims[0] * dstDims[1] * dstDims[2]
    outputArray = inputArray.NewInstance()
    outputArray.SetNumberOfTuples(newSize)
    offset = 0
    for k in range(dstDims[2]):
        for j in range(dstDims[1]):
            for i in range(dstDims[0]):
                srcIdx = (i * scale)
                srcIdx += (j * scale * srcDims[0])
                srcIdx += (k * scale * srcDims[0] * srcDims[1])
                outputArray.SetValue(offset, inputArray.GetValue(srcIdx))
                offset += 1

    return outputArray

# -----------------------------------------------------------------------------
# Image based exporter
# -----------------------------------------------------------------------------


def export_images(destinationPath, camera, **kwargs):
    # View size
    imageWidth = kwargs['imageWidth']
    imageHeight = kwargs['imageHeight']

    # Configure View
    view = simple.GetRenderView()
    view.ViewSize = [imageWidth, imageHeight]

    idb = ImageDataSetBuilder(destinationPath, 'image/jpg', camera)
    idb.start(view)
    idb.writeImages()
    idb.stop()

# -----------------------------------------------------------------------------
# Image based Volume exploration
# -----------------------------------------------------------------------------


def export_volume_exploration_images(destinationPath, camera, **kwargs):
    values = [int(v) for v in kwargs['multiValue'].split(',')]
    maxOpacity = float(kwargs['maxOpacity']) / 100.0
    span = float(kwargs['tentWidth']) * 0.5

    # View size
    imageWidth = kwargs['imageWidth']
    imageHeight = kwargs['imageHeight']

    # Configure View
    view = simple.GetRenderView()
    view.ViewSize = [imageWidth, imageHeight]

    pvw = get_volume_piecewise(view)
    if pvw:
        savedNodes = []
        currentPoints = [0, 0, 0, 0]
        for i in range(pvw.GetSize()):
            pvw.GetNodeValue(i, currentPoints)
            savedNodes.append([v for v in currentPoints])

        idb = ImageDataSetBuilder(destinationPath, 'image/jpg', camera)
        idb.getDataHandler().registerArgument(priority=1, name='volume',
                                              values=values, ui='slider',
                                              loop='reverse')
        idb.start(view)
        for volume in idb.getDataHandler().volume:
            pvw.RemoveAllPoints()
            pvw.AddPoint(float(volume) - span, 0)
            pvw.AddPoint(float(volume), maxOpacity)
            pvw.AddPoint(float(volume) + span, 0)
            pvw.AddPoint(255, 0)
            idb.writeImages()
        idb.stop()

        # Reset to original piecewise function
        pvw.RemoveAllPoints()
        for node in savedNodes:
            pvw.AddPoint(node[0], node[1], node[2], node[3])
    else:
        print('Can not export Volume exploration without a Volume')

# -----------------------------------------------------------------------------
# Image based Contour exploration
# -----------------------------------------------------------------------------


def export_contour_exploration_images(destinationPath, camera, **kwargs):
    values = [int(v) for v in kwargs['multiValue'].split(',')]

    # View size
    imageWidth = kwargs['imageWidth']
    imageHeight = kwargs['imageHeight']

    # Configure View
    view = simple.GetRenderView()
    view.ViewSize = [imageWidth, imageHeight]

    contour = get_contour()
    if contour:
        originalValues = [v for v in contour.Value]
        idb = ImageDataSetBuilder(destinationPath, 'image/jpg', camera)
        idb.getDataHandler().registerArgument(priority=1, name='contour',
                                              values=values, ui='slider',
                                              loop='reverse')
        idb.start(view)
        for contourValue in idb.getDataHandler().contour:
            contour.Value = [contourValue]
            idb.writeImages()
        idb.stop()

        # Reset to original value
        contour.Value = originalValues
    else:
        print('Can not export Contour exploration without a Contour.')

# -----------------------------------------------------------------------------
# Contours Geometry export
# -----------------------------------------------------------------------------


def export_contours_geometry(destinationPath, **kwargs):
    view = simple.GetRenderView()
    sceneDescription = {'scene': []}
    for key, value in simple.GetSources().items():
        if key[0] == 'Contour':
            add_scene_item(sceneDescription, key[0], value, view)

    count = 1
    for item in sceneDescription['scene']:
        item['name'] += ' (%d)' % count
        count += 1

    if count > 1:
        contour = sceneDescription['scene'][0]['source']
        sections = {
            'LookupTables': get_source_lookuptable_section(contour)
        }
        # Create geometry Builder
        dsb = VTKGeometryDataSetBuilder(destinationPath, sceneDescription,
                                        {}, sections)
        dsb.start()
        dsb.writeData(0)
        dsb.stop()

        # Patch data range
        patch_data_range(destinationPath)
    else:
        print('Can not export Contour(s) geometry without at least a Contour.')

# -----------------------------------------------------------------------------
# Contours Geometry export
# -----------------------------------------------------------------------------


def export_contour_exploration_geometry(destinationPath, **kwargs):
    values = [int(v) for v in kwargs['multiValue'].split(',')]
    contour = None
    for key, value in simple.GetSources().items():
        if key[0] == 'Contour':
            contour = value

    if contour:
        sections = {'LookupTables': get_source_lookuptable_section(contour)}
        scalarName = simple.GetRepresentation(contour).ColorArrayName[1]
        originalValue = [v for v in contour.Value]
        sceneDescription = {
            'scene': [{
                'name': 'Contour',
                'source': contour,
                'colors': {
                    scalarName: {
                        'constant': 0,
                        'location': 'POINT_DATA'
                    }
                }
            }]
        }
        dsb = VTKGeometryDataSetBuilder(destinationPath, sceneDescription,
                                        {}, sections)
        dsb.getDataHandler().registerArgument(priority=1, name='contour',
                                              values=values,
                                              ui='slider', loop='modulo')
        dsb.start()
        scalarContainer = sceneDescription['scene'][0]['colors'][scalarName]
        for contourValue in dsb.getDataHandler().contour:
            contour.Value = [contourValue]
            scalarContainer['constant'] = contourValue
            dsb.writeData()
        dsb.stop()

        # Patch data range
        patch_data_range(destinationPath)

        # Reset to original state
        contour.Value = originalValue
    else:
        print('Can not export Contour geometry without Contour(s)')


# -----------------------------------------------------------------------------
# Volume export
# -----------------------------------------------------------------------------


def export_volume(destinationPath, **kwargs):
    producer = get_trivial_producer()
    if not producer:
        return

    scale = int(kwargs['volumeScale'])
    view = simple.GetRenderView()
    arraName = producer.GetPointDataInformation().GetArray(0).Name
    indexJSON = {
        'type': ['tonic-query-data-model', 'vtk-volume'],
        'arguments': {},
        'arguments_order': [],
        'data': [{
            'rootFile': True,
            'name': 'scene',
            'pattern': 'volume.json',
            'priority': 0,
            'type': 'json',
            'metadata': {}
        }],
        'metadata': {}
    }
    volumeJSON = {
        'origin': [0, 0, 0],
        'spacing': [1, 1, 1],
        'extent': [0, 1, 0, 1, 0, 1],
        'vtkClass': 'vtkImageData',
        'pointData': {
            'vtkClass': 'vtkDataSetAttributes',
            'arrays': [{
                'data': {
                    'numberOfComponents': 1,
                    'name': arraName,
                    'vtkClass': 'vtkDataArray',
                    'dataType': 'Uint8Array',
                    'ref': {
                        'registration': 'setScalars',
                        'encode': 'LittleEndian',
                        'basepath': 'data',
                        'id': 'fieldData'
                    },
                    'size': 0
                }
            }]
        }
    }

    # Add color map
    indexJSON['LookupTables'] = get_volume_lookuptable_section(view)

    # create directories if need be
    dataDir = os.path.join(destinationPath, 'data')
    if not os.path.exists(dataDir):
        os.makedirs(dataDir)

    # Extract data
    imageData = producer.SMProxy.GetClientSideObject().GetOutputDataObject(0)
    extent = imageData.GetExtent()
    srcDims = (extent[1] - extent[0] + 1,
               extent[3] - extent[2] + 1,
               extent[5] - extent[4] + 1)
    dstDims = [int(v / scale) for v in srcDims]
    volumeJSON['extent'] = [0, dstDims[0] - 1,
                            0, dstDims[1] - 1,
                            0, dstDims[2] - 1]

    scalars = array_sampler(srcDims, dstDims, scale,
                            imageData.GetPointData().GetScalars())
    arraySize = scalars.GetNumberOfValues()
    volumeJSON['pointData']['arrays'][0]['data']['size'] = arraySize
    volumeJSON['pointData']['arrays'][0]['data']['dataType']  \
        = jsMapping[arrayTypesMapping[scalars.GetDataType()]]

    # Extract piecewise function
    pvw = get_volume_piecewise(view)
    if pvw:
        piecewiseNodes = []
        currentPoints = [0, 0, 0, 0]
        for i in range(pvw.GetSize()):
            pvw.GetNodeValue(i, currentPoints)
            piecewiseNodes.append([v for v in currentPoints])
        indexJSON['metadata'] = {'piecewise': piecewiseNodes}

    # Index file
    indexPath = os.path.join(destinationPath, 'index.json')
    with open(indexPath, 'w', encoding='utf8') as f:
        f.write(json.dumps(indexJSON, indent=2))

    # Image Data file
    volumePath = os.path.join(destinationPath, 'volume.json')
    with open(volumePath, 'w', encoding='utf8') as f:
        f.write(json.dumps(volumeJSON, indent=2))

    # Write data field
    fieldDataPath = os.path.join(destinationPath, 'data', 'fieldData')
    with open(fieldDataPath, 'wb') as f:
        f.write(memoryview(scalars))

# -----------------------------------------------------------------------------
# Composite exporter
# -----------------------------------------------------------------------------


def export_layers(destinationPath, camera):
    view = simple.GetRenderView()
    fp = tuple(view.CameraFocalPoint)
    cp = tuple(view.CameraPosition)
    vu = tuple(view.CameraViewUp)
    sceneDescription = {
        'size': tuple(view.ViewSize),
        'light': ['intensity'],  # 'normal', intensity
        'camera': {
            'CameraViewUp': vu,
            'CameraPosition': cp,
            'CameraFocalPoint': fp
        },
        'scene': []
    }

    for key, value in simple.GetSources().items():
        add_scene_item(sceneDescription, key[0], value, view)

    # Generate export
    dsb = CompositeDataSetBuilder(
        destinationPath, sceneDescription, camera, {}, {}, view)
    dsb.start()
    dsb.writeData()
    dsb.stop(compress=False)



def transform(dataset, size=2):
    """Apply a Median filter to dataset."""
    """ Median filter is a nonlinear filter used to reduce noise."""

    import scipy.ndimage
    import numpy as np

    array = dataset.active_scalars

    # Transform the dataset.
    result = np.empty_like(array)
    scipy.ndimage.filters.median_filter(array, size, output=result)

    # Set the result as the new scalars.
    dataset.active_scalars = result



from scipy.interpolate import interp1d
import numpy as np


class WBP:

    def __init__(self, Nx, Ny, Nz):

        self.Nslice = Nx
        self.Nray = Ny
        self.Nproj = Nz

        self.recon = np.empty([self.Nslice, self.Nray, self.Nray],
                              dtype=np.float32, order='F')

    def set_tilt_series(self, input_projections, input_tilt_angles):

        self.tiltSeries = input_projections
        self.tiltAngles = input_tilt_angles

        self.Nproj = self.tiltSeries.shape[2]

    def WBP(self, sliceInd):

        # Main Reconstruction Loop
        self.recon[sliceInd, :, :] = self.wbp2(self.tiltSeries[sliceInd, :, :])

    def wbp2(self, sinogram, filter="ramp", interp="linear"):

        ang = np.double(self.tiltAngles) * np.pi / 180.0
        # Create Fourier filter
        F = self.makeFilter(filter)
        # Pad sinogram for filtering
        s = np.lib.pad(sinogram, ((0, F.size - self.Nray), (0, 0)),
                       'constant', constant_values=(0, 0))
        # Apply Fourier filter
        s = np.fft.fft(s, axis=0) * F
        s = np.real(np.fft.ifft(s, axis=0))
        # Change back to original
        s = s[:self.Nray, :]

        # Back projection
        reconSlice = np.zeros((self.Nray, self.Nray), np.float32)
        center_proj = self.Nray // 2  # Index of center of projection
        [X, Y] = np.mgrid[0:self.Nray, 0:self.Nray]
        xpr = X - int(self.Nray) // 2
        ypr = Y - int(self.Nray) // 2

        for j in range(self.Nproj):
            t = ypr * np.cos(ang[j]) - xpr * np.sin(ang[j])
            x = np.arange(self.Nray) - center_proj
            if interp == 'linear':
                bp = np.interp(t, x, s[:, j], left=0, right=0)
            elif interp == 'spline':
                interpolant = interp1d(
                    x, s[:, j], kind='slinear', bounds_error=False,
                    fill_value=0)
                bp = interpolant(t)
            else:
                interpolant = interp1d(
                    x, s[:, j], kind=interp, bounds_error=False, fill_value=0)
                bp = interpolant(t)
            reconSlice = reconSlice + bp

        # Normalize
        reconSlice = reconSlice * np.pi / 2 / self.Nproj
        reconSlice[reconSlice < 0] = 0
        return reconSlice

    def makeFilter(self, filterMethod="ramp"):
        # Calculate next power of 2
        N2 = 2**np.ceil(np.log2(self.Nray))
        # Make a ramp filter.
        freq = np.fft.fftfreq(int(N2)).reshape(-1, 1)
        omega = 2 * np.pi * freq
        filter = 2 * np.abs(freq)

        if filterMethod == "ramp":
            pass
        elif filterMethod == "shepp-logan":
            filter[1:] = filter[1:] * np.sin(omega[1:]) / omega[1:]
        elif filterMethod == "cosine":
            filter[1:] = filter[1:] * np.cos(filter[1:])
        elif filterMethod == "hamming":
            filter[1:] = filter[1:] * (0.54 + 0.46 * np.cos(omega[1:] / 2))
        elif filterMethod == "hann":
            filter[1:] = filter[1:] * (1 + np.cos(omega[1:] / 2)) / 2
        elif filterMethod == "none":
            filter[:] = 1
        else:
            raise ValueError("Unknown filter: %s" % filterMethod)

        return filter



import tomviz.operators


class DefaultITKTransform(tomviz.operators.CancelableOperator):

    def transform(self, dataset):
        """Define this method for Python operators that transform the input
        array. This example uses an ITK filter to add 10 to each voxel value."""

        # Try imports to make sure we have everything that is needed
        try:
            from tomviz import itkutils
            import itk
        except Exception as exc:
            print("Could not import necessary module(s)")
            raise exc

        self.progress.value = 0
        self.progress.maximum = 100

        # Add a try/except around the ITK portion. ITK exceptions are
        # passed up to the Python layer, so we can at least report what
        # went wrong with the script, e.g., unsupported image type.
        try:
            self.progress.value = 0
            self.progress.message = "Converting data to ITK image"

            # Get the ITK image
            itk_image = itkutils.dataset_to_itk_image(dataset)
            itk_input_image_type = type(itk_image)
            self.progress.value = 30
            self.progress.message = "Running filter"

            # ITK filter
            filter = itk.AddImageFilter[itk_input_image_type, # Input 1
                                        itk_input_image_type, # Input 2
                                        itk_input_image_type].New() # Output
            filter.SetInput1(itk_image)
            filter.SetConstant2(10)
            itkutils.observe_filter_progress(self, filter, 30, 70)

            try:
                filter.Update()
            except RuntimeError: # Exception thrown when ITK filter is aborted
                return

            self.progress.message = "Saving results"

            itkutils.set_itk_image_on_dataset(filter.GetOutput(), dataset)

            self.progress.value = 100
        except Exception as exc:
            print("Problem encountered while running %s" %
                  self.__class__.__name__)
            raise exc



import json
import types

from marshmallow import (
    fields,
    Schema,
    post_load,
    EXCLUDE,
    INCLUDE,
    pre_load,
    post_dump
)

from ._models import (
    Tomviz,
    Pipeline,
    DataSource,
    Reader
)

from ._utils import (
    to_namespaces,
    from_namespace,
    attrs
)

from tomviz import operators, modules


class OperatorField(fields.Field):
    def _serialize(self, value, attr, obj, **kwargs):
        if value is None:
            return ""

        state = {}
        for a in dir(value):
            if a.startswith('_') or a == 'name':
                continue

            v = getattr(value, a)
            if a == 'dataSources':
                s = DataSourceSchema()
                v = [s.dump(d) for d in v]
                state[a] = v
            elif attrs(v):
                state[a] = from_namespace(v)
            else:
                state[a] = v

        return state

    def _deserialize(self, value, attr, data, **kwargs):
        from ._jsonpatch import update
        try:
            description = json.loads(value['description'])
            name = description['name']
        except (json.decoder.JSONDecodeError, KeyError):
            name = ''.join([x.capitalize() for x in value['label'].split(' ')])
        cls = getattr(operators, name)

        args = {}
        if 'arguments' in value:
            args['arguments'] = to_namespaces(value['arguments'])
        if 'dataSources' in value:
            s = DataSourceSchema(context=self.context)
            datasources = value['dataSources']
            datasources = [s.load(d) for d in datasources]
            args['dataSources'] = datasources
        args['id'] = value['id']

        removed_cache = self.context.get('operators', {})
        op = cls(**args)

        # Do we a op that we can recurrect?
        if op.id in removed_cache:
            op = update(op, removed_cache.pop(op.id))

        return op


class OperatorSchema(Schema):
    operator = OperatorField()


def load_operator(operator, removed_cache=None):
    schema = OperatorSchema()
    if removed_cache is not None:
        schema.context = removed_cache

    return schema.load({
        'operator': operator
    })['operator']


def dump_operator(op):
    ns = types.SimpleNamespace(operator=op)
    return OperatorSchema().dump(ns)['operator']


class ModuleField(fields.Field):
    def _serialize(self, value, attr, obj, **kwargs):
        if value is None:
            return ""

        state = {
            'type': value.__class__.__name__
        }
        for a in dir(value):
            if a.startswith('_'):
                continue

            v = getattr(value, a)
            if attrs(v):
                v = from_namespace(v)

            state[a] = v

        return state

    def _deserialize(self, value, attr, data, **kwargs):
        from ._jsonpatch import update
        cls = getattr(modules, value['type'])

        removed_cache = self.context.get('modules', {})
        mod = cls(**value)

        # Do we a module that we can recurrect?
        if mod.id in removed_cache:
            mod = update(mod, removed_cache.pop(mod.id))

        return mod


class ModuleSchema(Schema):
    module = ModuleField()


def load_module(module, removed_cache=None):
    schema = ModuleSchema()
    if removed_cache is not None:
        schema.context = removed_cache
    return schema.load({
        'module': module
    })['module']


def dump_module(module):
    ns = types.SimpleNamespace(module=module)
    return ModuleSchema().dump(ns)['module']


class ColorMap2DBoxSchema(Schema):
    x = fields.Float()
    y = fields.Float()
    height = fields.Float()
    width = fields.Float()

    @post_load
    def make_namespace(self, data, **kwargs):
        return to_namespaces(data)

    class Meta:
        unknown = INCLUDE


class ColorOpacityMap(Schema):
    colorSpace = fields.String()
    colors = fields.List(fields.Float)
    points = fields.List(fields.Float)

    @post_load
    def make_namespace(self, data, **kwargs):
        return to_namespaces(data)

    class Meta:
        unknown = INCLUDE


class GradientOpacityMap(Schema):
    points = fields.List(fields.Float)

    @post_load
    def make_namespace(self, data, **kwargs):
        return to_namespaces(data)

    class Meta:
        unknown = INCLUDE


class SubsampleSettingsSchema(Schema):
    # TODO: maybe we can force 'strides' to be of length 3,
    # and volumeBounds to be of length 6?
    strides = fields.List(fields.Int())
    volumeBounds = fields.List(fields.Int())


class ReaderSchema(Schema):
    fileNames = fields.List(fields.String)
    name = fields.String()
    subsampleSettings = fields.Nested(SubsampleSettingsSchema())

    @post_load
    def make_reader(self, data, **kwargs):
        return Reader(**data)

    @post_dump
    def remove_empty(self, data, **kwargs):
        if 'name' in data and data['name'] is None:
            del data['name']

        return data


class DataSourceSchema(Schema):
    operators = fields.List(OperatorField, missing=lambda: [])
    colorMap2DBox = fields.Nested(ColorMap2DBoxSchema)
    colorOpacityMap = fields.Nested(ColorOpacityMap)
    gradientOpacityMap = fields.Nested(GradientOpacityMap)
    modules = fields.List(ModuleField, missing=lambda: [])
    useDetachedColorMap = fields.Boolean()
    spacing = fields.List(fields.Float)
    units = fields.String()
    id = fields.String()
    label = fields.String()
    reader = fields.Nested(ReaderSchema)
    active = fields.Boolean()

    @post_load
    def make_datasource(self, data, **kwargs):
        from ._jsonpatch import update

        ds = DataSource(**data)
        # Has this data source been removed, if so recurrent it,
        # rather than creating a new one.
        removed_cache = self.context.get('dataSources', {})
        if ds.id in removed_cache:
            update(ds, removed_cache.pop(ds.id))

        return ds

    @post_dump
    def remove_empty(self, data, **kwargs):
        remove_if_empty = ['operators', 'modules']

        return {
            k: v for k, v in data.items() if k not in remove_if_empty or v != []
        }

    class Meta:
        unknown = EXCLUDE


def load_datasource(datasource, removed_cache=None):
    schema = DataSourceSchema()
    if removed_cache is not None:
        schema.context = removed_cache

    return DataSourceSchema().load(datasource)


def dump_datasource(datasource):
    return DataSourceSchema().dump(datasource)


class PipelineSchema(Schema):
    dataSource = fields.Nested(DataSourceSchema)

    # The following @pre_load and @post_dump are need to allow use to have
    # a Pipeline object with a datasource attribute. The JSON object in
    # the serialized state that represents a pipeline has not such attribute.
    # Using these two hooks allows use to achieve the object graph we want from
    # the serialize state.
    @pre_load
    def wrap_datasource(self, data, **kwargs):
        """
        Add the DataSource as an attribute of the pipeline.
        """
        return {
            'dataSource': data
        }

    @post_dump
    def unwrap_datasource(self, data, **kwargs):
        """
        Extract DataSource from pipeline attribute.
        """

        return data['dataSource']

    @post_load
    def make_pipeline(self, data, **kwargs):
        return Pipeline(data['dataSource'])

    class Meta:
        unknown = EXCLUDE


class TomvizSchema(Schema):
    pipelines = fields.List(fields.Nested(PipelineSchema),
                            data_key='dataSources')

    @post_load
    def make_tomviz(self, data, **kwargs):
        return Tomviz(data['pipelines'])

    class Meta:
        unknown = EXCLUDE



import numpy as np
import tomviz.operators


class AddPoissonNoiseOperator(tomviz.operators.CancelableOperator):

    def transform(self, dataset, N=25):
        """Add Poisson noise to tilt images"""
        self.progress.maximum = 1

        tiltSeries = dataset.active_scalars.astype(float)
        if tiltSeries is None:
            raise RuntimeError("No scalars found!")

        Ndata = tiltSeries.shape[0] * tiltSeries.shape[1]

        self.progress.maximum = tiltSeries.shape[2]
        step = 0
        for i in range(tiltSeries.shape[2]):
            if self.canceled:
                return

            tiltImage = tiltSeries[:, :, i].copy()
            tiltImage = tiltImage / np.sum(tiltSeries[:, :, i]) * (Ndata * N)
            tiltImage = np.random.poisson(tiltImage)
            tiltImage = tiltImage * np.sum(tiltSeries[:, :, i]) / (Ndata * N)

            tiltSeries[:, :, i] = tiltImage.copy()
            step += 1
            self.progress.value = step

        dataset.active_scalars = tiltSeries
